//========================================================================
//$Id: Buffers.java,v 1.1 2005/10/05 14:09:25 janb Exp $
//Copyright 2004-2005 Mort Bay Consulting Pty. Ltd.
//------------------------------------------------------------------------
//Licensed under the Apache License, Version 2.0 (the "License");
//you may not use this file except in compliance with the License.
//You may obtain a copy of the License at
//http://www.apache.org/licenses/LICENSE-2.0
//Unless required by applicable law or agreed to in writing, software
//distributed under the License is distributed on an "AS IS" BASIS,
//WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//See the License for the specific language governing permissions and
//limitations under the License.
//========================================================================


package org.mortbay.jetty.plugin.deb;


import java.io.*;
import java.util.List;
import java.util.Iterator;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugin.logging.Log;
import org.apache.maven.project.MavenProject;
import org.codehaus.plexus.util.FileUtils;
import org.codehaus.plexus.util.DirectoryScanner;
import org.codehaus.plexus.util.cli.Commandline;
import org.codehaus.plexus.util.cli.CommandLineUtils;

/**
 * Creates a Debian .deb Package File
 *
 * @author Leopoldo Lee Agdeppa III
 * @requiresDependencyResolution runtime
 * @goal deb
 * @phase package
 */
public class JettyDebMojo extends AbstractMojo
{
    private static String FILE_SEPARATOR = System.getProperty("file.separator");


    /**
     * Project instance, used to add new source directory to the build.
     *
     * @parameter default-value="${project}"
     * @required
     * @readonly
     */
    private MavenProject project;


    /**
     * @parameter
     * @optional
     */
    private String debFileName;

    // required fields Section


    /**
     * Debian Package Name: The value of this field determines the package name,
     * and is used to generate file names by most installation tools
     *
     * @parameter default-value="${project.artifactId}"
     * @required
     */
    private String packageName;


    /**
     * Debian Package Version: Typically, this is the original package's version number in
     * whatever form the program's author uses. It may also include a Debian revision number (for non-native packages).
     * If both version and revision are supplied, they are seperated by a hyphen, `-'.
     * For this reason, the original version may not have a hyphen in its version number
     *
     * @parameter default-value="${project.version}"
     * @required
     */
    private String version;




    /**
     * Debian Package Architecture: <arch|all>
     * The architecture specifies which type of hardware this package was compiled for.
     * Common architectures are `i386', `m68k', `sparc', `alpha', `powerpc' etc.
     * Note that the all option is meant for packages that are architecture independent.
     * Some examples of this are shell or Perl scripts, or documentation.
     *
     * @parameter default-value="all"
     * @optional
     */
    private String architecture;


    /**
     * Debian Package Maintainer's Name.
     *
     * @parameter
     * @required
     */
    private String maintainerName;

    /**
     * Debian Package Maintainer's Email.
     *
     * @parameter
     * @required
     */
    private String maintainerEmail;


    /**
     * Debian Package Short Description.
     *
     * @parameter
     * @required
     */
    private String shortDescription;


    /**
     * Debian Package Long Description:  more detailed description.
     * Each line of the long description must be preceded by a space,
     * and blank lines in the long desription must contain a single '.'
     * following the preceding space
     *
     * @parameter
     * @required
     */
    private String longDescription;

    // optional fields Section


    /**
     * Debian Package Section: This is a general field that gives the package a category based
     * on the software that it installs.
     * Some common sections are `utils', `net', `mail', `text', `x11' etc.
     *
     * @parameter
     * @optional
     */
    private String section;


    /**
     * Debian Package Priority: Sets the importance of this package in relation to the system as a whole.
     * Common priorities are `required', `standard', `optional', `extra' etc.
     *
     * @parameter
     * @optional
     */
    private String priority;


    /**
     * Debian Package Essential: <yes|no>
     * This field is usually only needed when the answer is `yes'.
     * It denotes a package that is required for proper operation of the system.
     * Dpkg or any other installation tool will not allow an Essential package to be
     * removed (at least not without using one of the force options).
     *
     * @parameter
     * @optional
     */
    private String essential;





    /**
     * Debian Package Source: <source name>
     * The name of the source package that this binary package came from,
     * if different than the name of the package itself.
     *
     * @parameter
     * @optional
     */
    private String source;


    /**
     * Debian Package Depends: <package list>
     * List of packages that are required for this package to provide a non-trivial amount of functionality.
     * The package maintenance software will not allow a package to be installed if
     * the packages listed in its Depends field aren't installed
     * (at least not without using the force options),
     * and will run the postinst scripts of packages listed in
     * Depends: fields before those of the packages which depend on them, and run prerm scripts before.
     *
     * @parameter
     * @optional
     */
    private String depends;


    /**
     * Debian Package Pre-Depends: <package list>
     * List of packages that must be installed and configured before this one can be installed.
     * This is usually used in the case where this package requires another package
     * for running its preinst script.
     *
     * @parameter
     */
    private String preDepends;

    /**
     * Debian Package Essentialk: <package list>
     * Lists packages that would be found together with this one in all but unusual installations.
     * The package maintenance software will warn the user if they install a package without those
     * listed in its Recommends field.
     *
     * @parameter
     * @optional
     */
    private String recomends;


    /**
     * Debian Package Suggests: <package list>
     * Lists packages that are related to this one and can perhaps enhance its usefulness,
     * but without which installing this package is perfectly reasonable.
     *
     * @parameter
     * @optional
     */
    private String suggests;


    /**
     * Debian Package installedSize.
     *
     * @parameter
     * @optional
     */
    private String installedSize;


    /**
     * Debian Package Conflicts: <package list>
     * Lists packages that conflict with this one,
     * for example by containing files with the same names.
     * The package maintenance software will not allow conflicting packages to be installed at the same time.
     * Two conflicting packages should each include a
     * Conflicts line mentioning the other.
     *
     * @parameter
     * @optional
     */
    private String conficts;


    /**
     * Debian Package Replaces: <package list>
     * List of packages from which this package is allowed to replace files.
     * This is used for allowing this package to overwrite the files of another package and is
     * usually used with the Conflicts field to force removal of the other package,
     * if this one also has the same files as the conflicted package.
     *
     * @parameter
     */
    private String replaces;


    /**
     * Debian Package Provides: <package list>
     * This is a list of virtual packages that this one provides.
     * Usuaully this is used in the case of several packages all
     * providing the same service. For example, sendmail and exim
     * can serve as a mail server, so they provide a common package (`mail-transport-agent')
     * on which other packages can depend. This will allow sendmail or exim to serve
     * as a valid option to satisfy the dependency. This prevents the packages that depend on a
     * mail server from having to know the package names for
     * all of them, and using `|' to separate the list.
     *
     * @parameter
     * @optional
     */
    private String provides;


    /**
     * The list of file-mappings.
     *
     * @parameter
     * @required
     */
    private List fileMappings;

    // Scripts


    /**
     * preinst: pre install script
     * This script executes before that package will be unpacked from its Debian archive (".deb") file.
     * Many 'preinst' scripts stop services for packages which are being upgraded until their installation
     * or upgrade is completed (following the successful execution of the 'postinst' script).
     *
     * @parameter
     * @optional
     */
    private String preinst;


    /**
     * postinst: post install script
     * This script typically completes any required configuration of the package foo once foo has been unpacked
     * from its Debian archive (".deb") file. Often, 'postinst' scripts ask the user for input, and/or warn the
     * user that if he accepts default values, he should remember to go back and re-configure that package as the
     * situation warrants. Many 'postinst' scripts then execute any commands necessary to start or restart a
     * service once a new package has been installed or upgraded
     *
     * @parameter
     * @optional
     */
    private String postinst;


    /**
     * prerm: pre remove script
     * This script typically stops any daemons which are associated with a package.
     * It is executed before the removal of files associated with the package.
     *
     * @parameter
     * @optional
     */
    private String prerm;


    /**
     * postrm: post remove script
     * This script typically modifies links or other files associated with foo,
     * and/or removes files created by the package.
     *
     * @parameter
     * @optional
     */
    private String postrm;


    /**
     * the fake root folder.
     *
     * @parameter expression="${project.build.directory}/debian"
     */
    private File fakeRoot;


    private File debianFolder;


    /**
     * Package File Name: Filename for the debian pacakage, example yourpackage.deb
     *
     * @parameter default-value="${project.artifactId}-${project.artifact.baseVersion}.deb"
     * @optional
     */
    private String packageFileName;


    public void execute() throws MojoExecutionException, MojoFailureException
    {
        
        validateConfigurations();

        initFolders();
        makeControlFile();
        createScripts();
        processFileMappings();


        try
        {
            executeDpkg();
        }
        catch (Exception e)
        {
            throw new MojoFailureException("Debian Package (dpkg --build) Tool Execution Error: " + e.getMessage());
        }


    }


    private void validateConfigurations() throws MojoExecutionException, MojoFailureException
    {


    }

    private void initFolders() throws MojoExecutionException, MojoFailureException
    {
        // Create the fake root folder
        if (!fakeRoot.exists())
        {
            getLog().info("Creating directory " + fakeRoot.getAbsolutePath());
            if (!fakeRoot.mkdirs())
            {
                throw new MojoFailureException("Unable to create directory " + fakeRoot.getAbsolutePath());
            }
        }

        // Create the DEBIAN Folder
        debianFolder = new File(fakeRoot, "DEBIAN");
        if (!debianFolder.exists())
        {
            getLog().info("Creating directory " + debianFolder.getAbsolutePath());
            if (!debianFolder.mkdir())
            {
                throw new MojoFailureException("Unable to create directory " + debianFolder.getAbsolutePath());
            }
        }

        try
        {
            chmod("775", debianFolder);
        }
        catch (Exception e)
        {
            throw new MojoFailureException("Unable to create directory " + debianFolder.getAbsolutePath());
        }
    }


    private void makeControlFile() throws MojoExecutionException, MojoFailureException
    {

        File control = new File(this.debianFolder, "control");

        // remove the file if it already exists
        if (control.exists())
        {
            control.delete();
        }


        try
        {
            getLog().info("Creating spec file " + control.getAbsolutePath());
            PrintWriter ctrl = new PrintWriter(new FileWriter(control));


            ctrl.println("Package: " + packageName.trim());
            ctrl.println("Version: " + version.trim());
            ctrl.println("Architecture: " + architecture.trim());
            ctrl.println("Maintainer: "
                    + maintainerName.trim()
                    + " "
                    +  (maintainerEmail != null ? (" <" + maintainerEmail.trim() + ">") : ""));


            if (section != null)
            {
                ctrl.println("Section: " + section.trim());
            }

            if (priority != null)
            {
                ctrl.println("Priority: " + priority.trim());
            }

            if (essential != null)
            {
                ctrl.println("Essential: " + essential.trim());
            }

            if (source != null)
            {
                ctrl.println("Source: " + source.trim());
            }

            if (depends != null)
            {
                ctrl.println("Depends: " + depends.trim());
            }

            if (preDepends != null)
            {
                ctrl.println("Pre-Depends: " + preDepends.trim());
            }

            if (recomends != null)
            {
                ctrl.println("Recomends: " + recomends.trim());
            }

            if (suggests != null)
            {
                ctrl.println("Suggests: " + suggests.trim());
            }

//            if(installedSize != null)
//            {
//                ctrl.println("Installed-Size: " + installedSize);
//            }


            if (conficts != null)
            {
                ctrl.println("Conficts: " + conficts.trim());
            }

            if (replaces != null)
            {
                ctrl.println("Replaces: " + replaces.trim());
            }

            if (provides != null)
            {
                ctrl.println("Provides: " + provides.trim());
            }


            ctrl.println("Description: " + shortDescription.trim());
            ctrl.println(" " + longDescription.trim().replaceAll("\n", " "));


            ctrl.close();

            chmod("775", control);
        }
        catch (Exception e)
        {
            e.printStackTrace();
            throw new MojoExecutionException("Unable to write " + control.getAbsolutePath(), e);
        }


    }


    private void createScripts() throws MojoExecutionException
    {
        createFile(new File(debianFolder, "preinst"), preinst);
        createFile(new File(debianFolder, "postinst"), postinst);
        createFile(new File(debianFolder, "prerm"), prerm);
        createFile(new File(debianFolder, "postrm"), postrm);
    }


    private void createFile(File file, String content) throws MojoExecutionException
    {
        try
        {
            if (file.exists())
            {
                file.delete();
            }

            if (content == null)
            {
                return;
            }


            PrintWriter f = new PrintWriter(new FileWriter(file));
            f.print(content);
            f.close();

            chmod("775", file);


        }
        catch (Exception e)
        {
            throw new MojoExecutionException("Install Script Generation Error: " +  e.getMessage());
        }
    }

    private void chmod(String mode, File file) throws Exception
    {

        Commandline cl = new Commandline();

        cl.setExecutable("chmod");
        String[] args = {mode, file.getAbsolutePath()};
        cl.addArguments(args);
        cl.setWorkingDirectory(debianFolder.getAbsolutePath());


        CommandLineUtils.StringStreamConsumer stdout = new CommandLineUtils.StringStreamConsumer();
        CommandLineUtils.StringStreamConsumer stderr = new CommandLineUtils.StringStreamConsumer();

        int code = CommandLineUtils.executeCommandLine(cl, stdout, stderr);

        System.out.println(stdout.getOutput());
        System.err.println(stderr.getOutput());
        if (code != 0)
        {
            throw new MojoExecutionException(stderr.getOutput());
        }



    }


    private void processFileMappings() throws MojoExecutionException
    {
        Iterator i = fileMappings.iterator();

        while (i.hasNext())
        {
            FileMapping mapping = (FileMapping) i.next();
            getLog().info(mapping.toString());
            copyFileMappingToFakeRoot(mapping);
        }

    }


    private void copyFileMappingToFakeRoot(FileMapping fileMapping) throws MojoExecutionException
    {
        File dest = new File(fakeRoot, fileMapping.getDestination());
        Log log = getLog();

        if (!fileMapping.getSource().exists())
        {
            throw new MojoExecutionException("Source Does not Exists: " + fileMapping.getSource().getAbsolutePath());
        }

        try
        {
            if (fileMapping.getSource().isDirectory())
            {
                copyFilesAndFolders(fileMapping);
            }
            else
            {
                FileUtils.copyFile(fileMapping.getSource(), dest);
            }
        }
        catch (IOException e)
        {
            e.printStackTrace();
            throw new MojoExecutionException("File or Folder Copy Error: source:"
                    + fileMapping.getSource().getAbsolutePath()
                    + " to Destination: "
                    + dest.getAbsolutePath(), e
            );
        }


    }


    private void copyFilesAndFolders(FileMapping fileMapping) throws IOException
    {

        Log log = getLog();

        DirectoryScanner scanner = new DirectoryScanner();
        scanner.setBasedir(fileMapping.getSource());
        scanner.setIncludes(fileMapping.getIncludesArray());
        scanner.setExcludes(fileMapping.getExcludesArray());
        scanner.setCaseSensitive(true);
        scanner.scan();


        String directories[] = scanner.getIncludedDirectories();
        String files[] = scanner.getIncludedFiles();
        String fr = fakeRoot.getCanonicalPath();
        String src;
        String dest;


        for (int i = 0; i < directories.length; i++)
        {
            dest = fr + FILE_SEPARATOR + fileMapping.getDestination() + FILE_SEPARATOR + directories[i];

            File file = new File(dest);
            if (!file.exists())
            {
                file.mkdir();
            }


        }


        for (int i = 0; i < files.length; i++)
        {
            src = fileMapping.getSource() + FILE_SEPARATOR + files[i];
            dest = fr + FILE_SEPARATOR + fileMapping.getDestination() + FILE_SEPARATOR + files[i];

            File srcFile = new File(src);
            File destFile = new File(dest);

            // copy only the modfied file and the none existing file
            if (destFile.lastModified() < srcFile.lastModified())
            {
                log.info("Add File To Deb: " + files[i]);
                FileUtils.copyFile(srcFile, destFile);

            }


        }


    }


    private void executeDpkg() throws MojoExecutionException, Exception
    {
        


        Commandline cl = new Commandline();

        cl.setExecutable("dpkg");
        File packageFile = new File(fakeRoot, packageFileName);
        String[] args = {"--build", fakeRoot.getCanonicalPath(), packageFile.getCanonicalPath()};
        cl.addArguments(args);
        cl.setWorkingDirectory(project.getBuild().getDirectory());

        
        CommandLineUtils.StringStreamConsumer stdout = new CommandLineUtils.StringStreamConsumer();
        CommandLineUtils.StringStreamConsumer stderr = new CommandLineUtils.StringStreamConsumer();

        int code = CommandLineUtils.executeCommandLine(cl, stdout, stderr);

        System.out.println(stdout.getOutput());
        System.err.println(stderr.getOutput());
        if (code != 0)
        {
            throw new MojoExecutionException(stderr.getOutput());
        }


    }


}
